/*;  dlist.cc {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.1 $$ {\rhead{\sl dlist.cc\quad\thepage}} */
/*;  $$Date: 1996/12/02 12:19:57 $$  */
/*;  $$Source: /home/shiiba/96prog/dlist/dlist.cc,v $$  */

#ifndef dlist_cc
#define dlist_cc

#include "dlist.h"

template <class T>
void DList<T>::DeepCopy
(
 const DList<T> & dlst
 )
{
  if (dlst.Count == 0)
    return;
  
  Node * n = dlst.Head;
  
  do  {
    Append(n->Data);
    n = n->Next;
  }
  while (n != NULL);
  
  Work = Head;
  Shallow  = NO;
  Circular = dlst.Circular;
}

// copy constructor
template <class T>
DList<T>::DList
(
 const DList<T> & dlst,
 Boolean shallow
 )
{
  if (shallow)
    ShallowCopy(dlst);
  else
    {
      SetNull();
      DeepCopy(dlst);
    }
}

// construct from array
template <class T>
DList<T>::DList		// 󤫤ʣ̤ƥꥹȤ
(
 const T * array,	// ʣ̸ؤΥݥ
 size_t no		// ˳ǼƤǤθĿ
 )
{
  SetNull();
  Circular = NO;
  Shallow  = NO;
  
  if ((array == NULL) || (no == 0))
    {
      error();  abort(); // ꤵƤʤǤ̵
    }
  
  const T * aptr = array;
  
  for (size_t n = 0; n < no; ++n)
    {
      Append(*aptr);
      ++aptr;
    }
}

// conversion to array operator
template <class T>
DList<T>::operator T * ()
{
  if (Count == 0)
    return NULL;
  
  T * result = new T [Count];
  
  if (result == NULL)
    {
      error(); abort();	// ҡפ˺Ǥʤä
    }
  
  Node * temp = Head;
  T    * aptr = result;
  
  while (temp != NULL)
    {
      *aptr = temp->Data;
      ++aptr;
      temp = temp->Next;
    }
  
  // note: caller must delete this array!
  return result;
}

// append new item
template <class T>
void DList<T>::Append
(
 const T & item
 )
{
  Node * n = new Node;
  
  if (n == NULL)
    {
      error(); abort();	// Ǥʤä
    }
  
  n->Data = item;
  n->Next = NULL;
  n->Prev = Tail;
  
  if (Count == 0)
    {
      Head = n;
      Tail = n;
      Work = n;
    }
  else
    {
      Tail->Next = n;
      Tail = n;
    }
  
  ++Count;
}

template <class T>
void DList<T>::Append
(
 const T * array,
 size_t no
 )
{
  if ((array == NULL) || (no == 0))
    {
      error(); abort();	// ɲä褦ȤƤ
    }
  
  const T * aptr = array;
  
  for (size_t n = 0; n < no; ++n)
    {
      Append(*aptr);
      ++aptr;
    }
}

template <class T>
void DList<T>::Append
(
 const DList<T> & dlst
 )
{
  const Node * n = dlst.Head;
  
  while (n != NULL)
    {
      Append(n->Data);
      n = n->Next;
    }
}

// delete current item
template <class T>
void DList<T>::Delete()
{
  if (Count == 0)
    {
      error(); abort();	// Ǥ̵Τˤ˾ääơ
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }
  
  if (Work == Head)
    {
      if (Work->Next == NULL)
	{
	  delete Work;
	  SetNull();
	}
      else
	{
	  Head = Work->Next;
	  Head->Prev = NULL;
	  delete Work;
	  Work = Head;
	}
    }
  else
    {
      if (Work == Tail)
	{
	  if (Work->Prev == NULL)
	    {
	      delete Work;
	      SetNull();
	    }
	  else
	    {
	      Tail = Work->Prev;
	      Tail->Next = NULL;
	      delete Work;
	      Work = Tail;
	    }
	}
      else
	{
	  Node * n = Work->Next;
	  Work->Prev->Next = Work->Next;
	  Work->Next->Prev = Work->Prev;
	  delete Work;
	  Work = n;
	}
    }
  
  if (Count > 0)
    --Count;
}

// insert new item in current location
template <class T>
void DList<T>::InsertBefore
(
 const T & item
 )
{
  if (Count == 0)
    {
      error(); abort();
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }
  
  Node * newnode = new Node;
  
  if (newnode == NULL)
    {
      error(); abort();
    }
  
  newnode->Data = item;
  
  if (Work == Head)
    {
      newnode->Next = Head;
      newnode->Prev = NULL;
      Head->Prev    = newnode;
      Head          = newnode;
    }
  else
    {
      Work->Prev->Next = newnode;
      newnode->Prev    = Work->Prev;
      newnode->Next    = Work;
      Work->Prev       = newnode;
    }
  
  ++Count;
}

template <class T>
void DList<T>::InsertBefore
(
 const T * array,
 size_t no
 )
{
  for (size_t n = 0; n < no; ++n)
    InsertBefore(array[n]);
}

template <class T>
void DList<T>::InsertBefore
(
 const DList<T> & dlst
 )
{
  const Node * n = dlst.Head;
  
  while (n != NULL)
    {
      InsertBefore(n->Data);
      n = n->Next;
    }
}

template <class T>
void DList<T>::InsertAfter
(
 const T & item
 )
{
  if (Count == 0)
    {
      error(); abort();
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }
  
  Node * newnode = new Node;
  
  if (newnode == NULL)
    {
      error(); abort();
    }
  
  newnode->Data = item;
  
  if (Work == Tail)
    {
      newnode->Next = NULL;
      newnode->Prev = Tail;
      Tail->Next    = newnode;
      Tail          = newnode;
    }
  else
    {
      Work->Next->Prev = newnode;
      newnode->Next    = Work->Next;
      newnode->Prev    = Work;
      Work->Next       = newnode;
    }
  
  ++Count;
}

template <class T>
void DList<T>::InsertAfter
(
 const T * array,
 size_t no
 )
{
  // insert in reverse order, so they're stored in order
  for (size_t n = no; n > 0; --n)
    InsertAfter(array[n-1]);
}

template <class T>
void DList<T>::InsertAfter
(
 const DList<T> & dlst
 )
{
  const Node * n = dlst.Tail;
  
  while (n != NULL)
    {
      InsertAfter(n->Data);
      n = n->Prev;
    }
}

// remove all items from the list
template <class T>
void DList<T>::Erase()
{
  if (!Shallow)
    {
      Work = Head;
      
      while (Work != NULL)
	{
	  Head = Work->Next;
	  delete Work;
	  Work = Head;
	}
    }
  
  SetNull();
  Shallow = NO;
}

// get current item
template <class T>
T DList<T>::Get()
{
  if (Count == 0)
    {
      error(); abort();
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }
  
  return Work->Data;
}

// directly reference current item
template <class T>
T & DList<T>::operator * ()
{
  if (Count == 0)
    {
      error(); abort();
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }

  return Work->Data;
}

// move to next item in list
template <class T>
Boolean DList<T>::ToNext()
{
  if (Count == 0)
    {
      error(); abort();
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }
  
  if (Work == Tail)
    {
      if (Circular)
	Work = Head;
      else
	return NO;
    }
  else
    Work = Work->Next;
  
  return YES;
}

template <class T>
Boolean DList<T>::ToPrev()
{
  if (Count == 0)
    {
      error(); abort();
    }
  
  if (Work == NULL)
    {
      error(); abort();
    }
  
  if (Work == Head)
    {
      if (Circular)
	Work = Tail;
      else
	return NO;
    }
  else
    Work = Work->Prev;
  
  return YES;
}

#endif

/*; end of file */
