/*;  gtool.cc {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.1 $$ {\rhead{\sl gtool.cc\quad\thepage}} */
/*;  $$Date: 1997/03/01 21:11:13 $$  */
/*;  $$Source: /home/sakakiba/96prog/slope/gtool.cc,v $$  */

#include "gtool.h"
#include "math.h"

// Area2 calculates the twice of the area of the triangle with points
// p0, p1 and p2, and return the value.
// If the area is positive, the direction of the points p0, p1 and p2 is 
// counter-clockwise. If the area is negative, the direction of the
// points p0, p1 and p2 is clockwise. If the area is zero, the vertices 
// are on the same straight line.

Metric Area2(Point2d p0, Point2d p1, Point2d p2) {
  Metric dx1, dx2, dy1, dy2;
  dx1 = p1.x() - p0.x(); dy1 = p1.y() - p0.y();
  dx2 = p2.x() - p0.x(); dy2 = p2.y() - p0.y();
  return (dx1*dy2 - dy1*dx2);
}

// Ccw finds out whether the direction of the points p0, p1, and p2 is 
// counter clockwise or not.
// The case when p0, p1, and p2 make a triangle:
//   return value     the direction of those three points
//   ------------     -----------------------------------
//      +1            counter clockwise
//      -1            clockwise
// The case when p0, p1, and p2 are on the same straight line:
//   return value     the order of those three points
//   ------------     -------------------------------
//      +1            p0, p1, p2
//       0            p0, p2, p1
//      -1            p2, p0, p1
// The original source code is found in "Algorithms in C++, Robert
// Sedgewick" 

int Ccw(Point2d p0, Point2d p1, Point2d p2) {
  Metric dx1, dx2, dy1, dy2;
  dx1 = p1.x() - p0.x(); dy1 = p1.y() - p0.y();
  dx2 = p2.x() - p0.x(); dy2 = p2.y() - p0.y();
  if (dx1*dy2 > dy1*dx2) return +1;
  if (dx1*dy2 < dy1*dx2) return -1;
  if ((dx1*dx2 < 0) || (dy1*dy2 < 0)) return -1;
  if ((dx1*dx1+dy1*dy1) < (dx2*dx2+dy2*dy2)) return +1;
  return 0;
}

// Intersect finds out whether the two line are intersects or not.
// When the two lines intersect, then the function returns 1.
// Otherwise the function returns 0.
// The original source code is found in "Algorithms in C++, Robert
// Sedgewick" 

int Intersect(Line2d Line1, Line2d Line2) {
  return ((Ccw(Line1.p1(), Line1.p2(), Line2.p1())
	   *Ccw(Line1.p1(), Line1.p2(), Line2.p2())) <= 0)
    && ((Ccw(Line2.p1(), Line2.p2(), Line1.p1())
	 *Ccw(Line2.p1(), Line2.p2(), Line1.p2())) <= 0);
}

// dist2 calculates the square of the length between two points

Metric dist2(Point2d p0, Point2d p1) {
  Metric dx = p0.x() - p1.x();
  Metric dy = p0.y() - p1.y();
  return (dx*dx + dy*dy);
}

// Inside finds out whether the poin is inside of the triangle or not.
// When the point is inside of the triangle, it returns 1.
// When the point is on a side of the triangle, it returns
// 0.
// When the point is outside of the triangle, it returns -1.

int Inside(Point2d pa, Point2d pb, Point2d pc, Point2d p) {
  
  int ca, cb, cc;
  if ((ca = Ccw(pa, pb, p)) == 0) return 0;
  if ((cb = Ccw(pb, pc, p)) == 0) return 0;
  if ((cc = Ccw(pc, pa, p)) == 0) return 0;
  if (fabs(ca + cb + cc) == 3.) return 1;
  return -1;
}

// Inside finds out whether the poin is inside of the quadrilateral or not.
// When the point is inside of the quadrilateral, it returns 1.
// When the point is on a side of the quadrilateral, it returns
// 0.
// When the point is outside of the quadrilateral, it returns -1.

int Inside(Point2d pa, Point2d pb, Point2d pc, Point2d pd, Point2d p) {
  
  int ca, cb, cc, cd;
  if ((ca = Ccw(pa, pb, p)) == 0) return 0;
  if ((cb = Ccw(pb, pc, p)) == 0) return 0;
  if ((cc = Ccw(pc, pd, p)) == 0) return 0;
  if ((cc = Ccw(pd, pa, p)) == 0) return 0;
  if (fabs(ca + cb + cc + cd) == 4.) return 1;
  return -1;
}

// CrossDiag finds the intersection of the diagonal lines of a
// quadrilateral.

Point3d CrossDiag(Point3d pa, Point3d pb, Point3d pc, Point3d pd) {
  
  double a1, b1, c1, a2, b2, c2;
  
  // line pa-pc: a1*x + b1*y = c1
  // line pb-pd: a2*x + b2*y = c2
  
  if (pa.x() == pc.x()) {
    a1 = 1.;    b1 = 0.;   c1 = pa.x();
  }
  else {
    b1 = 1.;
    a1 = - double(pc.y() - pa.y())/(pc.x() - pa.x());
    c1 = a1*pa.x() + b1*pa.y();
  }
  if (pb.x() == pd.x()) {
    a2 = 1.;    b2 = 0.;   c2 = pb.x();
  }
  else {
    b2 = 1.;
    a2 = - double(pd.y() - pb.y())/(pd.x() - pb.x());
    c2 = a2*pb.x() + b2*pb.y();
  }
  
  Point3d p;
  
  p.x(Metric( (c1*b2 - c2*b1)/(a1*b2 - a2*b1) + 0.5));
  p.y(Metric( (a1*c2 - a2*c1)/(a1*b2 - a2*b1) + 0.5));
  
  double da = sqrt(dist2(pa,p));
  double db = sqrt(dist2(pb,p));
  double dc = sqrt(dist2(pc,p));
  double dd = sqrt(dist2(pd,p));
  p.z(Metric( 0.5*(  (da*pc.z() + dc*pa.z())/(da + dc)
		    +(db*pd.z() + dd*pb.z())/(db + dd)) + 0.5));
  return p;
}	

// height caluculates the height of a point which is on the triangle

double
height(Point3d pa, Point3d pb, Point3d pc, Point2d p) {
  
  double det, deta, detb, detc;
  
  det =   pa.x()*pb.y() + pb.x()*pc.y() + pc.x()*pa.y()
    - pa.x()*pc.y() - pb.x()*pa.y() - pc.x()*pb.y();
  deta =  pa.z()*pb.y() + pb.z()*pc.y() + pc.z()*pa.y()
    - pa.z()*pc.y() - pb.z()*pa.y() - pc.z()*pb.y();
  detb =  pa.x()*pb.z() + pb.x()*pc.z() + pc.x()*pa.z()
    - pa.x()*pc.z() - pb.x()*pa.z() - pc.x()*pb.z();
  detc =  pa.x()*pb.y()*pc.z() + pb.x()*pc.y()*pa.z()
    + pc.x()*pa.y()*pb.z()
    - pa.x()*pc.y()*pb.z() - pb.x()*pa.y()*pc.z() - pc.x()*pb.y()*pa.z();
  
  return (deta*p.x() + detb*p.y() + detc)/det;
  
}

double
height(Point3d pa, Point3d pb, Point3d pc, Point3d pd, Point2d p) {
  
  Point3d pi = CrossDiag(pa, pb, pc, pd);
  
  if (Inside(pa, pb, pi, p) >= 0)
    return height(pa, pb, pi, p);
  else if (Inside(pb, pc, pi, p) >= 0)
    return height(pb, pc, pi, p);
  else if (Inside(pc, pd, pi, p) >= 0)
    return height(pc, pd, pi, p);
  else if (Inside(pd, pa, pi, p) >= 0)
    return height(pd, pa, pi, p);
  else
    return 0;
}

//; Rccs returns the position of a point relative to a line segment.

//; P_UP              in the upper side of the line p0-p1
//; P_DOWN            in the lower side of the line p0-p1
//; P_ONLINE          on the line p0-p1 including the both end
//;                   points 
//; P_RIGHT           to the right of the line p0-p1
//; P_LEFT            to the left of the line p0-p1

RelPosition
Rccw(			//; Find the relative position of a
			//; point p with  
			//; reference to a line segment
  Point2d p0,		//; the left end of a line segment
  Point2d p1,		//; the right end of a line segment
  Point2d p		//; the point to be checked
			//; return value: the relative position
			//; of p 
  )
{
  Metric dx1, dx2, dy1, dy2;
  dx1 = p1.x() - p0.x(); dy1 = p1.y() - p0.y();
  dx2 = p.x() - p0.x();  dy2 = p.y() - p0.y();
  if (dx1*dy2 > dy1*dx2) return P_UP;
  if (dx1*dy2 < dy1*dx2) return P_DOWN;
  if ((dx1*dx2 < 0) || (dy1*dy2 < 0)) return P_LEFT;
  if ((dx1*dx1+dy1*dy1) < (dx2*dx2+dy2*dy2)) return P_RIGHT;
  return P_ONLINE;
}

ostream& operator<<(
  ostream& s,
  RelPosition& pos
  )
{
  if (pos == P_LEFT) {
    s << "LEFT"; 
  }
  else if (pos ==P_ONLINE) {
    s << "ONLINE";
  }
  else if (pos == P_RIGHT) {
    s << "RIGHT"; 
  }
  else if (pos == P_UP) {
    s << "UP";
  }
  else if (pos == P_DOWN) {
    s << "DOWN";
  }
  return s;
}

Point2d
IntersectPoint(		// 뤳ȤʬäƤʬθζ᤯
			// ޤ
  Line2d Line1,		// ʬ1
  Line2d Line2		// ʬ1

  //  Intersect Ȥäơ뤳ȤϳǧƤΤȤޤ

  )
{
  double a1, b1, c1, a2, b2, c2;
  
  // Line1: a1*x + b1*y = c1
  // Line2: a2*x + b2*y = c2
  
  if (Line1.p1().x() == Line1.p2().x()) {
    a1 = 1.;    b1 = 0.;   c1 = Line1.p1().x();
  }
  else {
    b1 = 1.;
    a1 = - double(Line1.p2().y() - Line1.p1().y())
      /(Line1.p2().x() - Line1.p1().x());
    c1 = a1*Line1.p1().x() + b1*Line1.p1().y();
  }
  if (Line2.p1().x() == Line2.p2().x()) {
    a2 = 1.;    b2 = 0.;   c2 = Line2.p1().x();
  }
  else {
    b2 = 1.;
    a2 = - double(Line2.p2().y() - Line2.p1().y())
      /(Line2.p2().x() - Line2.p1().x());
    c2 = a2*Line2.p1().x() + b2*Line2.p1().y();
  }
  
  Point2d p;
  
  p.x(Metric( (c1*b2 - c2*b1)/(a1*b2 - a2*b1) + 0.5));
  p.y(Metric( (a1*c2 - a2*c1)/(a1*b2 - a2*b1) + 0.5));

  return p;
  
}

/*
float 
Theta(			// ʬ p1p2 ȿʿľȤʤ٤ޤ
  Point2d& p1,		// ʬ1
  Point2d& p2		// ʬ2
)
{
//  int dx, dy, ax, ay;
  Metric dx, dy, ax, ay;
  float t;
  dx = p2.x() - p1.x();  ax = (int)fabs(dx);
  dy = p2.y() - p1.y();  ay = (int)fabs(dy);
  t = (ax + ay == 0) ? 0 : (float) dy / (ax + ay);
  if (dx < 0)  t = 2 - t;
  else if (dy < 0) t = 4 + t;
  return t * 90.;
}
*/
  
/*; end of file */
